import time
import random
from pylibftdi import Device

# min, max, home
BOUNDS = {
    0: [1250, 1950, 1600],
    1: [850, 2500, 1000],
    # 2: [1200, 2500, 1200],
    2: [550, 1750, 600],
    3: [700, 2500, 1900],  # 1700 is center
    # 4: [1200, 2450, 1200],
}


class Arm(object):
    def __init__(self, fps, velocity_scale=200, dry_run=False, verbose=False):
        """
        fps: the frames per second that set_multi_velocity must be called to
        keep smooth motion.
        velocity_scale: convert unitless range (-1, 1) to velocity in
        microseconds / second
        """
        if not dry_run:
            self.dev = Device(mode='t')
            self.dev.baudrate = 9600

        self.default_speed = 600
        self.positions = self._home_position()

        self.fps = fps
        self.velocity_scale = velocity_scale

        # position scale is the velocity in (microseconds / second) / FPS to get
        # microseconds per frame
        self.position_scale = velocity_scale / self.fps

        self.dry_run = dry_run
        self.verbose = verbose

    def _home_position(self):
        return {i: BOUNDS[i][2] for i in BOUNDS}

    def _bound_position(self, axis, position):
        if position > BOUNDS[axis][1]:
            return BOUNDS[axis][1]
        if position < BOUNDS[axis][0]:
            return BOUNDS[axis][0]
        return position

    def set_position(self, axis, position, speed=None, time=None):
        """
        pos: position pulse width in microseconds
        speed: microseconds per second
        time: time in milliseconds to execute motion to `pos`
        """

        self.positions[axis] = self._bound_position(axis, position)

        if speed is None and time is None:
            speed = self.default_speed

        if self.verbose:
            print('axis=', axis)
            print('position=', position)
            print('speed=', speed)
            print('time=', time)

        if self.dry_run:
            return

        if speed:
            self.dev.write(
                '#{axis}P{position}S{speed}\r'.format(
                    axis=axis, position=position, speed=speed
                )
            )
        else:
            self.dev.write(
                '#{axis}P{position}T{time}\r'.format(
                    axis=axis, position=position, time=time
                )
            )

    def set_positions(self, positions, speeds=None, scaled=False):
        if scaled:
            positions = {
                axis: self._scaled_to_absoltuion_position(axis, position)
                for axis, position in positions.items()
            }

        for axis in positions:
            positions[axis] = self._bound_position(axis, positions[axis])
            self.positions[axis] = positions[axis]

        if speeds is None:
            speeds = {axis: self.default_speed for axis in positions}

        if self.verbose:
            print('positions', positions)
            print('speeds   ', speeds)

        if self.dry_run:
            return

        self.dev.write(
            ''.join(
                '#{axis}P{pos}S{speed}'.format(
                    axis=axis,
                    pos=positions[axis],
                    speed=speeds[axis],
                ) for axis in positions
            ) + '\r'
        )

    def _scaled_position(self, axis, position):
        return (position - BOUNDS[axis][0]) / (
            BOUNDS[axis][1] - BOUNDS[axis][0]
        )

    def scaled_positions(self):
        return [
            self._scaled_position(axis, position)
            for axis, position in self.positions.items()
        ]

    def _scaled_to_absoltuion_position(self, axis, position):
        if position < 0 or position > 1:
            raise ValueError((
                'position expected to be within 0 and 1.  found: {}'
            ).format(position))

        return BOUNDS[axis][0] + position * (BOUNDS[axis][1] - BOUNDS[axis][0])

    def set_scaled_position(self, axis, position, speed=None):
        self.set_position(
            axis,
            self._scaled_to_absoltuion_position(axis, position),
            speed=speed
        )

    def set_relative_position(self, axis, position_delta, speed=None):
        self.positions[axis] += position_delta
        self.set_position(axis, self.positions[axis], speed=speed)

    def set_velocities(self, velocities):
        """
        Set velocity of all servos in arm.

        set_multi_velocity must be called once every self.fps
        """
        if set(velocities.keys()) != set(self.positions.keys()):
            raise ValueError((
                'velocities.keys must match self.positions.keys:\n'
                '  velocities.keys(): {}\n'
                '  self.position.keys(): {}\n'
            ).format(velocities.keys(), self.positions.keys()))

        if not any(v != 0 for v in velocities.values()):
            return

        self.set_positions(
            {
                axis: self.positions[axis] + (velocity * self.position_scale)
                for axis, velocity in velocities.items()
            },
            {
                axis: max(abs(velocity * self.velocity_scale), 100)
                for axis, velocity in velocities.items()
            },
        )

    def set_velocity(self, axis, velocity):
        velocity *= 10
        self.set_position(
            axis,
            self.positions[axis] + velocity * self.position_scale,
            speed=max(abs(velocity) * self.velocity_scale, 5)
        )

    def go_home(self):
        self.set_positions(self._home_position())

    def go_random(self):
        for axis in BOUNDS:
            self.set_position(
                axis, random.randrange(BOUNDS[axis][0], BOUNDS[axis][1], 1)
            )


if __name__ == '__main__':
    arm = Arm()

    # dev.write('#0P500S400\r')
    # time.sleep(3)

    # dev.write('#0P1600S400\r')
    # dev.write('#1P1000S400\r')
    # dev.write('#2P1200S400\r')
    # dev.write('#3P1500S400\r')
    # dev.write('#4P1200S400\r')

    arm.go_home()

    try:
        a = [BOUNDS[i][2] for i in range(len(BOUNDS))]

        import sys
        while True:
            c = sys.stdin.read(1)
            if c == 'a':
                a[0] += 50
                arm.set_position(0, a[0], 200)
            elif c == 'z':
                a[0] -= 50
                arm.set_position(0, a[0], 200)

    except KeyboardInterrupt as e:
        pass

    arm.go_home()



"""
Sample Python/Pygame Programs
Simpson College Computer Science
http://programarcadegames.com/
http://simpson.edu/computer-science/

Show everything we can pull off the joystick
"""
import pygame
import time
import math
from ssc_32u.arm import Arm

FPS = 10
DRY_RUN = False

pygame.init()

# Set the width and height of the screen [width,height]
# size = [500, 700]
# screen = pygame.display.set_mode(size)

pygame.display.set_caption("My Game")

# Loop until the user clicks the close button.
done = False

# Used to manage how fast the screen updates
clock = pygame.time.Clock()

# Initialize the joysticks
pygame.joystick.init()

# Get ready to print
arm = Arm(FPS, dry_run=DRY_RUN, verbose=True, velocity_scale=1000)
arm.go_home()

# is this entire mapping something ROS would help with?
BUTTON_GO_HOME = 9
# map from joystick axes to arm axes
JOYSTICK_AXES_MAP = {
    0: 3,
    1: 0,
    3: -2,
    4: 1,
}
# map from joystick axes to logical axes
LOGICAL_AXES_MAP = {
    0: 0,
    1: 1,
    3: 2,
    4: 3,
}


def sign(x):
    if sign > 0:
        return 1
    elif sign == 0:
        return 0
    else:
        return -1


def interact_with_arm():
    # apply this to all joysticks for now ...
    for joystick_id in range(pygame.joystick.get_count()):
        joystick = pygame.joystick.Joystick(joystick_id)
        joystick.init()

        arm.set_velocities({
            LOGICAL_AXES_MAP[joystick_axis]: (
                joystick.get_axis(joystick_axis) *
                sign(arm_axis)
            )
            for joystick_axis, arm_axis in JOYSTICK_AXES_MAP.items()
        })

        # start button triggers go home
        buttons = joystick.get_numbuttons()
        for button_id in range(buttons):
            if button_id == BUTTON_GO_HOME and joystick.get_button(button_id):
                arm.go_home()


def display_joystick_status():
    # Get count of joysticks
    joystick_count = pygame.joystick.get_count()

    print("Number of joysticks: {}".format(joystick_count))

    # For each joystick:
    for i in range(joystick_count):
        joystick = pygame.joystick.Joystick(i)
        joystick.init()

        print("Joystick {}".format(i))

        # Get the name from the OS for the controller/joystick
        name = joystick.get_name()
        print("Joystick name: {}".format(name))

        # Usually axis run in pairs, up/down for one, and left/right for
        # the other.
        axes = joystick.get_numaxes()
        print("Number of axes: {}".format(axes))

        for i in range(axes):
            axis_value = joystick.get_axis(i)
            print("Axis {} value: {:>6.3f}".format(i, axis_value))

        buttons = joystick.get_numbuttons()
        print("Number of buttons: {}".format(buttons))

        for i in range(buttons):
            button = joystick.get_button(i)
            print("Button {:>2} value: {}".format(i, button))

        # Hat switch. All or nothing for direction, not like joysticks.
        # Value comes back in an array.
        hats = joystick.get_numhats()
        print("Number of hats: {}".format(hats))

        for i in range(hats):
            hat = joystick.get_hat(i)
            print("Hat {} value: {}".format(i, str(hat)))


steps = 0
old_clock = time.clock()


def display_measured_fps():
    global steps
    global old_clock

    steps += 1
    if steps % 100 == 0:
        new_clock = time.clock()
        print(100 / (new_clock - old_clock))
        old_clock = new_clock


# -------- Main Program Loop -----------
while not done:
    # EVENT PROCESSING STEP
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            done = True

        # Possible joystick actions: JOYAXISMOTION JOYBALLMOTION JOYBUTTONDOWN
        # JOYBUTTONUP JOYHATMOTION
        if event.type == pygame.JOYBUTTONDOWN:
            print("Joystick button pressed.")
        if event.type == pygame.JOYBUTTONUP:
            print("Joystick button released.")

    # display_joystick_status()
    interact_with_arm()

    # Limit to 60 frames per second
    clock.tick(FPS)

    display_measured_fps()

# Close the window and quit.
# If you forget this line, the program will 'hang'
# on exit if running from IDLE.
pygame.quit()



from setuptools import setup


setup(
    name='ssc_32u',
    version='0.2.2',
    description='simple interface to robot arms '
                'controlled with SSC-32U controller',
    url='http://github.com/dwiel/ssc_32u',
    author='Zach Dwiel',
    author_email='zdwiel@gmail.com',
    license='Apache',
    packages=['ssc_32u'],
    install_requires=[
        'pylibftdi',
    ],
)



"""
Sample Python/Pygame Programs
Simpson College Computer Science
http://programarcadegames.com/
http://simpson.edu/computer-science/

Show everything we can pull off the joystick
"""
import pygame
import time
import math
from ssc_32u.arm import Arm
from evdev import InputDevice, list_devices, ecodes, categorize, events

FPS = 10
DRY_RUN = False

# Loop until the user clicks the close button.
done = False

# Used to manage how fast the screen updates
clock = pygame.time.Clock()

# Initialize the joysticks
devices = [InputDevice(fn) for fn in list_devices()]
assert len(devices) == 1
dev = devices[0]

# Get ready to print
arm = Arm(FPS, dry_run=DRY_RUN, verbose=True, velocity_scale=1000)
arm.go_home()

# is this entire mapping something ROS would help with?
BUTTON_GO_HOME = 9
# map from joystick axes to (arm_axis, sign)
JOYSTICK_AXES_MAP = {
    0: (3, 1),
    1: (0, 1),
    3: (2, -1),
    4: (1, 1),
}


def interact_with_arm(velocities):
    arm.set_velocities({
        abs(arm_axis): velocities.get(joystick_axis, 0) * sign
        for joystick_axis, (arm_axis, sign) in JOYSTICK_AXES_MAP.items()
    })

    # # start button triggers go home
    # buttons = joystick.get_numbuttons()
    # for button_id in range(buttons):
    #     if button_id == BUTTON_GO_HOME and joystick.get_button(button_id):
    #         arm.go_home()


steps = 0
old_clock = time.clock()


def display_measured_fps():
    global steps
    global old_clock

    steps += 1
    if steps % 100 == 0:
        new_clock = time.clock()
        print(100 / (new_clock - old_clock))
        old_clock = new_clock

print(dev.capabilities(verbose=True))

# -------- Main Program Loop -----------
import threading


class JoystickThread(threading.Thread):
    def __init__(self):
        super(JoystickThread, self).__init__()

        # {joystick_axis: velocity (-1.0, 1.0)}
        self.state = {}

    def run(self):
        for event in dev.read_loop():
            time_delta = event.timestamp() - time.time()

            if time_delta > 0.1:
                print("warning: time delay: %.04f" % time_delta)

            if event.type == 0:
                if all(v > -3/128.0 and v < 3/128.0 for v in self.state.values()):
                    continue
            elif event.type == 3:
                self.state[event.code] = (event.value - 128) / 128.0

            event = categorize(event)
            if isinstance(event, events.KeyEvent):
                if event.keystate == events.KeyEvent.key_down:
                    if event.event.code == 297:
                        arm.go_home()


joystick = JoystickThread()
joystick.daemon = True
joystick.start()

while True:
    # send state
    print('send', joystick.state)
    interact_with_arm(joystick.state)

    # limit frames per second
    clock.tick(FPS)

    display_measured_fps()
